#include <iostream>
#include <fstream>
#include <string>
#include <stdio.h>
#include <cstdint>
#include <thread>
#include <chrono>
#include <atomic>
#include <dw/egomotion/Egomotion.h>
#include <dw/egomotion/global/GlobalEgomotion.h>
#include <dw/core/Context.h>
#include <dw/core/VersionCurrent.h>

using u64 = std::uint64_t;
using s64 = std::int64_t;

namespace sensor::odometry
{
	struct wheel_state
	{
		float radius;
		float rotation_speed;
		float contact_point_speed;
		u64 timestamp_us;
	};

	struct state
	{
		struct
		{
			wheel_state rear_left;
			wheel_state rear_right;
			wheel_state front_left;
			wheel_state front_right;
		} wheels;

		struct
		{
			float steering_angle;
			u64 timestamp_us;
		} steering;
	};
}

namespace dw
{
	call(const dwStatus status, const char* message)
	{
		if (status != DW_SUCCESS)
		{
			std::cerr << std::string(message) << std::endl;
			exit();
		}
	}
}

int main(int argc, char** argv)
{
	// intialize driveworks
	dwContextHandle_t context;
	dw::call(dwInitialize(&context, DW_VERSION, nullptr), "dwInitialize error");
	dwRigObject* rig_handle;
	dw::call(dwRig_initializeFromFile(&rig_handle, context, "data/rig.json"), "could not create rig from file");
	// initialize egomotion and global egomotion
	u32 gps_sensor_id;
	dw::call(dwRig_findSensorByName(&gps_sensor_id, "gps:rtk", rig_handle), "gps sensor id");
	dwEgomotionParameters egomotion_params{};
	dw::call(dwEgomotion_initParamsFromRig(&egomotion_params, rig_handle, nullptr, nullptr, nullptr), "init egomotion params from rig");
	egomotion_params.motionModel = DW_EGOMOTION_ODOMETRY;
	egomotion_params.automaticUpdate = true;

	dwEgomotionObject* egomotion;
	dw::call(dwEgomotion_initialize(&egomotion, &egomotion_params, context), "egomotion initialize");

	dwGlobalEgomotionParameters global_egomotion_params{};
	global_egomotion_params.historySize = 1000;
	global_egomotion_params.rotationalDrift = 10.0f;
	dwTransformation3f gps_to_rig{};
	dw::call(dwRig_getSensorToRigTransformation(&gps_to_rig, gps_sensor_id, rig_handle), "gps sensor to rig transformation");
	global_egomotion_params.sensorCharacteristics.antennaPosition.x = gps_to_rig.array[12];
	global_egomotion_params.sensorCharacteristics.antennaPosition.y = gps_to_rig.array[13];
	global_egomotion_params.sensorCharacteristics.antennaPosition.z = gps_to_rig.array[14];
	dwGlobalEgomotionObject* global_egomotion;
	dw::call(dwGlobalEgomotion_initialize(&global_egomotion, &global_egomotion_params, context), "global egomotion initialize");

	// ODOMETRY
	std::atomic<bool> odometry_done = false;
	std::thread odometry_thread([&odometry_done, &egomotion]
	{
		std::ifstream in_odo("data/odometry_data.bin");
		sensor::odometry::state odometry;
		in_odo.read(reinterpret_cast<char*>(&odometry), sizeof(sensor::odometry::state));
		sensor::odometry::state next_odo;
		while (in_odo.read(reinterpret_cast<char*>(&next_odo), sizeof(sensor::odometry::state)) && !odometry_done)
		{
			dwVehicleIOState vehicle_io_state{};
			vehicle_io_state.wheelSpeedTimestamp[DW_VEHICLE_WHEEL_REAR_LEFT] = odometry.wheels.rear_left.timestamp_us;
			vehicle_io_state.wheelSpeed[DW_VEHICLE_WHEEL_REAR_LEFT] = odometry.wheels.rear_left.rotation_speed;
			vehicle_io_state.wheelSpeedTimestamp[DW_VEHICLE_WHEEL_REAR_RIGHT] = odometry.wheels.rear_right.timestamp_us;
			vehicle_io_state.wheelSpeed[DW_VEHICLE_WHEEL_REAR_RIGHT] = odometry.wheels.rear_right.rotation_speed;
			vehicle_io_state.wheelSpeedTimestamp[DW_VEHICLE_WHEEL_FRONT_LEFT] = odometry.wheels.front_left.timestamp_us;
			vehicle_io_state.wheelSpeed[DW_VEHICLE_WHEEL_FRONT_LEFT] = odometry.wheels.front_left.rotation_speed;
			vehicle_io_state.wheelSpeedTimestamp[DW_VEHICLE_WHEEL_FRONT_RIGHT] = odometry.wheels.front_right.timestamp_us;
			vehicle_io_state.wheelSpeed[DW_VEHICLE_WHEEL_FRONT_RIGHT] = odometry.wheels.front_right.rotation_speed;
			vehicle_io_state.steeringTimestamp = odometry.steering.timestamp_us;
			vehicle_io_state.steeringAngle = odometry.steering.steering_angle;

			dw::call(dwEgomotion_addVehicleState(&vehicle_io_state, egomotion), "add vehicle state");

			std::this_thread::sleep_for(std::chrono::microseconds(next_odo.steering.timestamp_us - odometry.steering.timestamp_us));
			odometry = next_odo;
		}
		odometry_done = true;
	});

	// GPS
	std::atomic<bool> gps_done = false;
	std::thread gps_thread([&gps_done, &global_egomotion]
	{
		std::ifstream in_gps("data/gps_data.bin");
		dwGPSFrame gps;
		in_gps.read(reinterpret_cast<char*>(&gps), sizeof(dwGPSFrame));
		dwGPSFrame next_gps;
		while (in_gps.read(reinterpret_cast<char*>(&next_gps), sizeof(dwGPSFrame)) && !gps_done)
		{
			dw::call(dwGlobalEgomotion_addGPSMeasurement(&gps, global_egomotion), "add gps measurement");
			std::this_thread::sleep_for(std::chrono::microseconds(next_gps.timestamp_us - gps.timestamp_us));
			gps = next_gps;
		}
		gps_done = true;
	});

	while (!odometry_done && !gps_done)
	{
		std::this_thread::sleep_for(std::chrono::milliseconds(100));
		dwEgomotionResult estimate;
		dwEgomotionUncertainty uncertainty;
		if (dwEgomotion_getEstimation(&estimate, egomotion) == DW_SUCCESS &&
			dwEgomotion_getUncertainty(&uncertainty, egomotion) == DW_SUCCESS)
		{
			uncertainty.validFlags |= DW_EGOMOTION_ROTATION | DW_EGOMOTION_LIN_VEL_X | DW_EGOMOTION_ANG_VEL_Z; // I had to add these due else it didn't accept relative motion
			dw::call(dwGlobalEgomotion_addRelativeMotion(&estimate, &uncertainty, global_egomotion), "add relative motion");
		}

		dwGlobalEgomotionResult global_estimate;
		dwGlobalEgomotionUncertainty global_uncertainty;
		dw::call(dwGlobalEgomotion_getEstimate(&global_estimate, &global_uncertainty, global_egomotion), "get global estimate");

		// NO changes to global estimate !
	}
	odometry_done = true;
	gps_done = true;

	odometry_thread.join();
	gps_thread.join();

	return 0;
}
